@@ -1,3 +1,4 @@
+import sys
 from array import array
 
 try:
@@ -23,13 +24,9 @@ try:
 except ImportError:
     NumericType = None
 
+from decimal import Decimal
 from types import ClassType, InstanceType, NoneType
 
-try:
-    from decimal import Decimal
-except ImportError:
-    Decimal = None
-
 ########################################
 ## Quoting
 ########################################
@@ -154,20 +151,12 @@ def SequenceConverter(value, db):
 registerConverter(tuple, SequenceConverter)
 registerConverter(list, SequenceConverter)
 registerConverter(dict, SequenceConverter)
-try:
-    set, frozenset
-except NameError:
-    pass
-else:
-    registerConverter(set, SequenceConverter)
-    registerConverter(frozenset, SequenceConverter)
-try:
-    from sets import Set, ImmutableSet
-except ImportError:
-    pass
-else:
-    registerConverter(Set, SequenceConverter)
-    registerConverter(ImmutableSet, SequenceConverter)
+registerConverter(set, SequenceConverter)
+registerConverter(frozenset, SequenceConverter)
+if sys.version_info[:3] < (2, 6, 0): # Module sets was deprecated in Python 2.6
+   from sets import Set, ImmutableSet
+   registerConverter(Set, SequenceConverter)
+   registerConverter(ImmutableSet, SequenceConverter)
 
 if hasattr(time, 'struct_time'):
     def StructTimeConverter(value, db):
@@ -192,11 +181,11 @@ def TimeConverter(value, db):
 
 registerConverter(datetime.time, TimeConverter)
 
-if Decimal:
-    def DecimalConverter(value, db):
-        return value.to_eng_string()
+def DecimalConverter(value, db):
+    # See http://mail.python.org/pipermail/python-dev/2008-March/078189.html
+    return str(value.to_eng_string()) # Convert to str to work around a bug in Python 2.5.2
 
-    registerConverter(Decimal, DecimalConverter)
+registerConverter(Decimal, DecimalConverter)
 
 def sqlrepr(obj, db=None):
     try:
