@@ -2,7 +2,6 @@ import sqlbuilder
 NoDefault = sqlbuilder.NoDefault
 import styles
 import classregistry
-from col import popKey
 import events
 
 __all__ = ['MultipleJoin', 'SQLMultipleJoin', 'RelatedJoin', 'SQLRelatedJoin',
@@ -19,10 +18,7 @@ class Join(object):
     def __init__(self, otherClass=None, **kw):
         kw['otherClass'] = otherClass
         self.kw = kw
-        if self.kw.has_key('joinMethodName'):
-            self._joinMethodName = popKey(self.kw, 'joinMethodName')
-        else:
-            self._joinMethodName = None
+        self._joinMethodName = self.kw.pop('joinMethodName', None)
 
     def _set_joinMethodName(self, value):
         assert self._joinMethodName == value or self._joinMethodName is None, "You have already given an explicit joinMethodName (%s), and you are now setting it to %s" % (self._joinMethodName, value)
@@ -62,7 +58,7 @@ class SOJoin(object):
             otherClass, self._setOtherClass)
         self.joinColumn = joinColumn
         self.joinMethodName = joinMethodName
-        self.orderBy = orderBy
+        self._orderBy = orderBy
         if not self.joinColumn:
             # Here we set up the basic join, which is
             # one-to-many, where the other class points to
@@ -70,6 +66,12 @@ class SOJoin(object):
             self.joinColumn = styles.getStyle(
                 self.soClass).tableReference(self.soClass.sqlmeta.table)
 
+    def orderBy(self):
+        if self._orderBy is NoDefault:
+            self._orderBy = self.otherClass.sqlmeta.defaultOrder
+        return self._orderBy
+    orderBy = property(orderBy)
+
     def _setOtherClass(self, cls):
         self.otherClass = cls
 
@@ -77,14 +79,12 @@ class SOJoin(object):
         return False
 
     def _applyOrderBy(self, results, defaultSortClass):
-        if self.orderBy is NoDefault:
-            self.orderBy = defaultSortClass.sqlmeta.defaultOrder
         if self.orderBy is not None:
             results.sort(sorter(self.orderBy))
         return results
 
 def sorter(orderBy):
-    if isinstance(orderBy, tuple) or isinstance(orderBy, list):
+    if isinstance(orderBy, (tuple, list)):
         if len(orderBy) == 1:
             orderBy = orderBy[0]
         else:
@@ -166,8 +166,6 @@ class SOSQLMultipleJoin(SOMultipleJoin):
             conn = None
         pythonColumn = self._dbNameToPythonName()
         results = self.otherClass.select(getattr(self.otherClass.q, pythonColumn) == inst.id, connection=conn)
-        if self.orderBy is NoDefault:
-            self.orderBy = self.otherClass.sqlmeta.defaultOrder
         return results.orderBy(self.orderBy)
 
 class SQLMultipleJoin(Join):
@@ -242,6 +240,9 @@ class OtherTableToJoin(sqlbuilder.SQLExpression):
         self.interTable = interTable
         self.joinColumn = joinColumn
 
+    def tablesUsedImmediate(self):
+        return [self.otherTable, self.interTable]
+
     def __sqlrepr__(self, db):
         return '%s.%s = %s.%s' % (self.otherTable, self.otherIdName, self.interTable, self.joinColumn)
 
@@ -252,6 +253,9 @@ class JoinToTable(sqlbuilder.SQLExpression):
         self.interTable = interTable
         self.joinColumn = joinColumn
 
+    def tablesUsedImmediate(self):
+        return [self.table, self.interTable]
+
     def __sqlrepr__(self, db):
         return '%s.%s = %s.%s' % (self.interTable, self.joinColumn, self.table, self.idName)
 
@@ -261,6 +265,9 @@ class TableToId(sqlbuilder.SQLExpression):
         self.idName = idName
         self.idValue = idValue
 
+    def tablesUsedImmediate(self):
+        return [self.table]
+
     def __sqlrepr__(self, db):
         return '%s.%s = %s' % (self.table, self.idName, self.idValue)
 
@@ -282,8 +289,6 @@ class SOSQLRelatedJoin(SORelatedJoin):
             TableToId(self.soClass.sqlmeta.table, self.soClass.sqlmeta.idName, inst.id),
         ), clauseTables=(self.soClass.sqlmeta.table, self.otherClass.sqlmeta.table, self.intermediateTable),
         connection=conn)
-        if self.orderBy is NoDefault:
-            self.orderBy = self.otherClass.sqlmeta.defaultOrder
         return results.orderBy(self.orderBy)
 
 class SQLRelatedJoin(RelatedJoin):
@@ -295,7 +300,7 @@ def capitalize(name):
 class SOSingleJoin(SOMultipleJoin):
 
     def __init__(self, **kw):
-        self.makeDefault = popKey(kw, 'makeDefault', False)
+        self.makeDefault = kw.pop('makeDefault', False)
         SOMultipleJoin.__init__(self, **kw)
 
     def performJoin(self, inst):
