@@ -1,5 +1,5 @@
 """
-SQLObject 0.10.2
+SQLObject 0.12.4
 ----------------
 
 :author: Ian Bicking <ianb@colorstudy.com>
@@ -7,7 +7,7 @@ SQLObject 0.10.2
 SQLObject is a object-relational mapper.  See SQLObject.html or
 SQLObject.txt for more.
 
-With the help by Oleg Broytmann <phd@phd.pp.ru> and many other contributors.
+With the help by Oleg Broytman <phd@phd.pp.ru> and many other contributors.
 See Authors.txt.
 
 This program is free software; you can redistribute it and/or modify
@@ -610,6 +610,7 @@ class sqlmeta(object):
 
     def getColumns(sqlmeta):
         return sqlmeta.columns.copy()
+    getColumns = classmethod(getColumns)
 
     def asDict(self):
         """
@@ -657,9 +658,9 @@ class _sqlmeta_attr(object):
 warnings_level = 1
 exception_level = None
 # Current levels:
-#  1) Actively deprecated in version after 0.9; removed after
-#  2) Deprecated after 1 (0.9?)
-#  3) Deprecated after 2 (0.10?)
+#  1) Actively deprecated
+#  2) Deprecated after 1
+#  3) Deprecated after 2
 
 def deprecated(message, level=1, stacklevel=2):
     if exception_level is not None and exception_level <= level:
@@ -667,8 +668,8 @@ def deprecated(message, level=1, stacklevel=2):
     if warnings_level is not None and warnings_level <= level:
         warnings.warn(message, DeprecationWarning, stacklevel=stacklevel)
 
-#if sys.version_info[:3] < (2, 4, 0):
-#    deprecated("Support for Python 2.3 has been declared obsolete and will be removed in the next release of SQLObject")
+#if sys.version_info[:3] < (2, 5, 0):
+#    deprecated("Support for Python 2.4 has been declared obsolete and will be removed in the next release of SQLObject")
 
 def setDeprecationLevel(warning=1, exception=None):
     """
@@ -765,7 +766,10 @@ class SQLObject(object):
             if hasattr(mod, '__connection__'):
                 connection = mod.__connection__
 
-        if connection and not hasattr(cls, '_connection'):
+        # Do not check hasattr(cls, '_connection') here - it is possible
+        # SQLObject parent class has a connection attribute that came
+        # from sqlhub, e.g.; check __dict__ only.
+        if connection and ('_connection' not in cls.__dict__):
             cls.setConnection(connection)
 
         # We have to check if there are columns in the inherited
@@ -806,11 +810,10 @@ class SQLObject(object):
         # more.
         if not is_base:
             cls.q = sqlbuilder.SQLObjectTable(cls)
+            cls.j = sqlbuilder.SQLObjectTableWithJoins(cls)
 
         classregistry.registry(cls.sqlmeta.registry).addClass(cls)
 
-    _expired = _sqlmeta_attr('expired', 1)
-
     # @classmethod
     def _SO_setupSqlmeta(cls, new_attrs, is_base):
         """
@@ -861,7 +864,7 @@ class SQLObject(object):
         been deprecated; they are moved to the sqlmeta class, and
         a deprecation warning is given.
         """
-        for attr in ['_expired']:
+        for attr in ():
             if new_attrs.has_key(attr):
                 deprecated("%r is deprecated and read-only; please do "
                            "not use it in your classes until it is fully "
@@ -886,7 +889,7 @@ class SQLObject(object):
         if val is None:
             try:
                 val = cls(_SO_fetch_no_create=1)
-                val._SO_validatorState = SQLObjectState(val)
+                val._SO_validatorState = sqlbuilder.SQLObjectState(val)
                 val._init(id, connection, selectResults)
                 cache.put(id, cls, val)
             finally:
@@ -917,7 +920,8 @@ class SQLObject(object):
         # If no connection was given, we'll inherit the class
         # instance variable which should have a _connection
         # attribute.
-        if connection is not None:
+        if (connection is not None) and \
+                (getattr(self, '_connection', None) is not connection):
             self._connection = connection
             # Sometimes we need to know if this instance is
             # global or tied to a particular connection.
@@ -990,6 +994,11 @@ class SQLObject(object):
         finally:
             self._SO_writeLock.release()
 
+        post_funcs = []
+        self.sqlmeta.send(events.RowUpdatedSignal, self, post_funcs)
+        for func in post_funcs:
+            func()
+
     def expire(self):
         if self.sqlmeta.expired:
             return
@@ -1042,8 +1051,13 @@ class SQLObject(object):
         if self.sqlmeta.cacheValues:
             setattr(self, instanceName(name), value)
 
-    def set(self, **kw):
-        if not self.sqlmeta._creating and not getattr(self.sqlmeta, "row_update_sig_suppress", False):
+        post_funcs = []
+        self.sqlmeta.send(events.RowUpdatedSignal, self, post_funcs)
+        for func in post_funcs:
+            func()
+
+    def set(self, _suppress_set_sig=False, **kw):
+        if not self.sqlmeta._creating and not getattr(self.sqlmeta, "row_update_sig_suppress", False) and not _suppress_set_sig:
             self.sqlmeta.send(events.RowUpdateSignal, self, kw)
         # set() is used to update multiple values at once,
         # potentially with one SQL statement if possible.
@@ -1129,6 +1143,11 @@ class SQLObject(object):
         finally:
             self._SO_writeLock.release()
 
+        post_funcs = []
+        self.sqlmeta.send(events.RowUpdatedSignal, self, post_funcs)
+        for func in post_funcs:
+            func()
+
     def _SO_selectInit(self, row):
         for col, colValue in zip(self.sqlmeta.columnList, row):
             if col.to_python:
@@ -1188,9 +1207,10 @@ class SQLObject(object):
 
             # Pass the connection object along if we were given one.
             if kw.has_key('connection'):
-                self._connection = kw['connection']
-                self.sqlmeta._perConnection = True
-                del kw['connection']
+                connection = kw.pop('connection')
+                if getattr(self, '_connection', None) is not connection:
+                    self._connection = connection
+                    self.sqlmeta._perConnection = True
 
             self._SO_writeLock = threading.Lock()
 
@@ -1219,7 +1239,7 @@ class SQLObject(object):
 
         self.sqlmeta._creating = True
         self._SO_createValues = {}
-        self._SO_validatorState = SQLObjectState(self)
+        self._SO_validatorState = sqlbuilder.SQLObjectState(self)
 
         # First we do a little fix-up on the keywords we were
         # passed:
@@ -1288,16 +1308,18 @@ class SQLObject(object):
         return getID(obj)
 
     def _findAlternateID(cls, name, dbName, value, connection=None):
-        if isinstance(value, unicode):
-            column = cls.sqlmeta.columns[name]
-            if isinstance(column, col.SOUnicodeCol):
-                value = value.encode(column.dbEncoding)
         if isinstance(name, str):
             name = (name,)
             value = (value,)
         if len(name) != len(value):
             raise ValueError, "'column' and 'value' tuples must be of the same size"
-        condition = sqlbuilder.AND(*[getattr(cls.q, n)==v for n,v in zip(name, value)])
+        new_value = []
+        for n, v in zip(name, value):
+            from_python = getattr(cls, '_SO_from_python_' + n)
+            if from_python:
+                v = from_python(v, sqlbuilder.SQLObjectState(cls))
+            new_value.append(v)
+        condition = sqlbuilder.AND(*[getattr(cls.q, n)==v for n,v in zip(name, new_value)])
         return (connection or cls._connection)._SO_selectOneAlt(
             cls,
             [cls.sqlmeta.idName] +
@@ -1346,11 +1368,6 @@ class SQLObject(object):
     select = classmethod(select)
 
     def selectBy(cls, connection=None, **kw):
-        for key, column in cls.sqlmeta.columns.items():
-            if (key in kw) and isinstance(column, col.SOUnicodeCol):
-                value = kw[key]
-                if isinstance(value, unicode):
-                    kw[key] = value.encode(column.dbEncoding)
         conn = connection or cls._connection
         return cls.SelectResultsClass(cls,
                              conn._SO_columnClause(cls, kw),
@@ -1565,6 +1582,11 @@ class SQLObject(object):
         for func in post_funcs:
             func()
 
+        post_funcs = []
+        self.sqlmeta.send(events.RowDestroyedSignal, self, post_funcs)
+        for func in post_funcs:
+            func()
+
     def delete(cls, id, connection=None):
         obj = cls.get(id, connection=connection)
         obj.destroySelf()
@@ -1629,6 +1651,9 @@ class SQLObject(object):
     def tablesUsedImmediate(self):
         return [self.__class__.q]
 
+
+    # Comparison
+
     def __eq__(self, other):
         if self.__class__ is other.__class__:
             if self.id == other.id:
@@ -1666,13 +1691,6 @@ def instanceName(name):
     return '_SO_val_%s' % name
 
 
-class SQLObjectState(object):
-
-    def __init__(self, soObject):
-        self.soObject = weakref.proxy(soObject)
-        self.protocol = 'sql'
-
-
 ########################################
 ## Utility functions (for external consumption)
 ########################################
