@@ -1,23 +1,24 @@
 #!/usr/bin/env python
-import optparse
+
 import fnmatch
-import re
+import optparse
 import os
+import re
 import sys
 import textwrap
-import warnings
-try:
-    from paste import pyconfig
-    from paste import CONFIG
-except ImportError, e:
-    pyconfig = None
-    CONFIG = {}
 import time
+import warnings
+
+try:
+    from paste.deploy import appconfig
+except ImportError:
+    appconfig = None
 
 import sqlobject
 from sqlobject import col
-from sqlobject.util import moduleloader
+from sqlobject.classregistry import findClass
 from sqlobject.declarative import DeclarativeMeta
+from sqlobject.util import moduleloader
 
 # It's not very unsafe to use tempnam like we are doing:
 warnings.filterwarnings(
@@ -183,12 +184,87 @@ class Command(object):
 
     help = ''
 
+    def orderClassesByDependencyLevel(self, classes):
+        """
+        Return classes ordered by their depth in the class dependency
+        tree (this is *not* the inheritance tree), from the
+        top level (independant) classes to the deepest level.
+        The dependency tree is defined by the foreign key relations.
+        """
+        # @@: written as a self-contained function for now, to prevent
+        # having to modify any core SQLObject component and namespace
+        # contamination.
+        # yemartin - 2006-08-08
+
+        class SQLObjectCircularReferenceError(Exception): pass
+
+        def findReverseDependencies(cls):
+            """
+            Return a list of classes that cls depends on. Note that
+            "depends on" here mean "has a foreign key pointing to".
+            """
+            depended = []
+            for col in cls.sqlmeta.columnList:
+                if col.foreignKey:
+                    other = findClass(col.foreignKey,
+                                      col.soClass.sqlmeta.registry)
+                    if other not in depended:
+                        depended.append(other)
+            return depended
+
+        # Cache to save already calculated dependency levels.
+        dependency_levels = {}
+        def calculateDependencyLevel(cls, dependency_stack=[]):
+            """
+            Recursively calculate the dependency level of cls, while
+            using the dependency_stack to detect any circular reference.
+            """
+            # Return value from the cache if already calculated
+            if dependency_levels.has_key(cls):
+                return dependency_levels[cls]
+            # Check for circular references
+            if cls in dependency_stack:
+                dependency_stack.append(cls)
+                raise SQLObjectCircularReferenceError, (
+                        "Found a circular reference: %s " %
+                        (' --> '.join([x.__name__
+                                       for x in dependency_stack])))
+            dependency_stack.append(cls)
+            # Recursively inspect dependent classes.
+            depended = findReverseDependencies(cls)
+            if depended:
+                level = max([calculateDependencyLevel(x, dependency_stack)
+                             for x in depended]) + 1
+            else:
+                level = 0
+            dependency_levels[cls] = level
+            return level
+
+        # Now simply calculate and sort by dependency levels:
+        try:
+            sorter = []
+            for cls in classes:
+                level = calculateDependencyLevel(cls)
+                sorter.append((level, cls))
+            sorter.sort()
+            ordered_classes = [cls for level, cls in sorter]
+        except SQLObjectCircularReferenceError, msg:
+            # Failsafe: return the classes as-is if a circular reference
+            # prevented the dependency levels to be calculated.
+            print ("Warning: a circular reference was detected in the "
+                    "model. Unable to sort the classes by dependency: they "
+                    "will be treated in alphabetic order. This may or may "
+                    "not work depending on your database backend. "
+                    "The error was:\n%s" % msg)
+            return classes
+        return ordered_classes
+
     def __classinit__(cls, new_args):
         if cls.__bases__ == (object,):
             # This abstract base class
             return
         register(cls)
-    
+
     def __init__(self, invoked_as, command_name, args, runner):
         self.invoked_as = invoked_as
         self.command_name = command_name
@@ -287,7 +363,7 @@ class Command(object):
             else:
                 print 'No eggs specified'
             sys.exit(1)
-        return all
+        return self.orderClassesByDependencyLevel(all)
 
     def classes_from_module(self, module):
         all = []
@@ -320,13 +396,15 @@ class Command(object):
     def config(self):
         if not getattr(self.options, 'config_file', None):
             return None
-        if pyconfig and self.options.config_fn.endswith('.conf'):
-            config = pyconfig.Config(with_default=True)
-            config.load(self.options.config_file)
-            CONFIG.push_process_config(config)
-            return config
+        config_file = self.options.config_file
+        if appconfig:
+            if (not config_file.startswith('egg:')
+                and not config_file.startswith('config:')):
+                config_file = 'config:' + config_file
+            return appconfig(config_file,
+                             relative_to=os.getcwd())
         else:
-            return self.ini_config(self.options.config_file)
+            return self.ini_config(config_file)
 
     def ini_config(self, conf_fn):
         conf_section = 'main'
@@ -394,7 +472,7 @@ class Command(object):
                     continue
                 classes = self.classes_from_module(module)
                 all.extend(classes)
-                    
+
         os.path.walk(package_dir, find_classes_in_file, None)
         return all
 
@@ -419,15 +497,14 @@ class Command(object):
             dir = conf['history_dir']
             dir = dir.replace('$base', dist.location)
             self.options.output_dir = dir
-        
+
     def config_from_egg(self, egg_spec, warn_no_sqlobject=True):
         import pkg_resources
-        pkg_resources.require(egg_spec)
-        dist = pkg_resources.working_set.find(pkg_resources.Requirement(egg_spec))
+        dist = pkg_resources.get_distribution(egg_spec)
         if not dist.has_metadata('sqlobject.txt'):
             if warn_no_sqlobject:
                 print 'No sqlobject.txt in %s egg info' % egg_spec
-            return {}
+            return None, {}
         result = {}
         for line in dist.get_metadata_lines('sqlobject.txt'):
             line = line.strip()
@@ -436,7 +513,7 @@ class Command(object):
             name, value = line.split('=', 1)
             name = name.strip().lower()
             if name in result:
-                print 'Warning: %s appears more than one in sqlobject.txt' % name
+                print 'Warning: %s appears more than once in sqlobject.txt' % name
             result[name.strip().lower()] = value.strip()
         return dist, result
 
@@ -502,11 +579,20 @@ class CommandSQL(Command):
 
     def command(self):
         classes = self.classes()
+        allConstraints = []
         for cls in classes:
             if self.options.verbose >= 1:
                 print '-- %s from %s' % (
                     cls.__name__, cls.__module__)
-            print cls.createTableSQL().strip() + ';\n'
+            createSql, constraints = cls.createTableSQL()
+            print createSql.strip() + ';\n'
+            allConstraints.append(constraints)
+        for constraints in allConstraints:
+            if constraints:
+                for constraint in constraints:
+                    if constraint:
+                        print constraint.strip() + ';\n'
+
 
 class CommandList(Command):
 
@@ -540,14 +626,23 @@ class CommandCreate(Command):
         created = 0
         existing = 0
         dbs_created = []
+        constraints = {}
         for soClass in self.classes(require_some=True):
             if (self.options.create_db
                 and soClass._connection not in dbs_created):
                 if not self.options.simulate:
-                    soClass._connection.createEmptyDatabase()
+                    try:
+                        soClass._connection.createEmptyDatabase()
+                    except soClass._connection.module.ProgrammingError, e:
+                        if str(e).find('already exists') != -1:
+                            print 'Database already exists'
+                        else:
+                            raise
                 else:
                     print '(simulating; cannot create database)'
                 dbs_created.append(soClass._connection)
+            if soClass._connection not in constraints.keys():
+                constraints[soClass._connection] = []
             exists = soClass._connection.tableExists(soClass.sqlmeta.table)
             if v >= 1:
                 if exists:
@@ -556,22 +651,34 @@ class CommandCreate(Command):
                 else:
                     print 'Creating %s' % soClass.__name__
             if v >= 2:
-                print soClass.createTableSQL()
+                sql, extra = soClass.createTableSQL()
+                print sql
             if (not self.options.simulate
                 and not exists):
                 if self.options.interactive:
                     if self.ask('Create %s' % soClass.__name__):
                         created += 1
-                        soClass.createTable()
+                        tableConstraints = soClass.createTable(applyConstraints=False)
+                        if tableConstraints:
+                            constraints[soClass._connection].append(tableConstraints)
                     else:
                         print 'Cancelled'
                 else:
                     created += 1
-                    soClass.createTable()
+                    tableConstraints = soClass.createTable(applyConstraints=False)
+                    if tableConstraints:
+                        constraints[soClass._connection].append(tableConstraints)
+        for connection in constraints.keys():
+            if v >= 2:
+                print 'Creating constraints'
+            for constraintList in constraints[connection]:
+                for constraint in constraintList:
+                    if constraint:
+                        connection.query(constraint)
         if v >= 1:
             print '%i tables created (%i already exist)' % (
                 created, existing)
-        
+
 
 class CommandDrop(Command):
 
@@ -584,7 +691,7 @@ class CommandDrop(Command):
         v = self.options.verbose
         dropped = 0
         not_existing = 0
-        for soClass in self.classes():
+        for soClass in reversed(self.classes()):
             exists = soClass._connection.tableExists(soClass.sqlmeta.table)
             if v >= 1:
                 if exists:
@@ -769,11 +876,11 @@ class CommandRecord(Command):
     help = ('Record state of table definitions.  The state of each '
             'table is written out to a separate file in a directory, '
             'and that directory forms a "version".  A table is also '
-            'added to you datebase (%s) that reflects the version the '
+            'added to your database (%s) that reflects the version the '
             'database is currently at.  Use the upgrade command to '
             'sync databases with code.'
             % SQLObjectVersionTable.sqlmeta.table)
-               
+
     parser = standard_parser()
     parser.add_option('--output-dir',
                       help="Base directory for recorded definitions",
@@ -816,7 +923,7 @@ class CommandRecord(Command):
         if self.options.force_db_version:
             self.command_force_db_version()
             return
-        
+
         v = self.options.verbose
         sim = self.options.simulate
         classes = self.classes()
@@ -837,6 +944,12 @@ class CommandRecord(Command):
                               + '_' + dbName + '.sql')
             if sim:
                 continue
+            create, constraints = cls.createTableSQL()
+            if constraints:
+                constraints = '\n-- Constraints:\n%s\n' % (
+                    '\n'.join(constraints))
+            else:
+                constraints = ''
             files[fn] = ''.join([
                 '-- Exported definition from %s\n'
                 % time.strftime('%Y-%m-%dT%H:%M:%S'),
@@ -844,8 +957,9 @@ class CommandRecord(Command):
                 % (cls.__module__, cls.__name__),
                 '-- Database: %s\n'
                 % dbName,
-                cls.createTableSQL().strip(),
-                '\n'])
+                create.strip(),
+                '\n',
+                constraints])
         last_version_dir = self.find_last_version()
         if last_version_dir and not self.options.force_create:
             if v > 1:
@@ -894,25 +1008,25 @@ class CommandRecord(Command):
                 f = open(os.path.join(output_dir, fn), 'w')
                 f.write(content)
                 f.close()
-        all_diffs = []
-        for cls in self.classes():
-            for conn in conns:
-                diffs = db_differences(cls, conn)
-                for diff in diffs:
-                    if len(conns) > 1:
-                        diff = '  (%s).%s: %s' % (
-                            conn.uri(), cls.sqlmeta.table, diff)
-                    else:
-                        diff = '  %s: %s' % (cls.sqlmeta.table, diff)
-                    all_diffs.append(diff)
-        if all_diffs:
-            print 'Database does not match schema:'
-            print '\n'.join(all_diffs)
-            if self.options.db_record:
-                print '(Not updating database version)'
-        elif self.options.db_record:
-            for conn in conns:
-                self.update_db(version, conn)
+        if self.options.db_record:
+            all_diffs = []
+            for cls in self.classes():
+                for conn in conns:
+                    diffs = db_differences(cls, conn)
+                    for diff in diffs:
+                        if len(conns) > 1:
+                            diff = '  (%s).%s: %s' % (
+                                conn.uri(), cls.sqlmeta.table, diff)
+                        else:
+                            diff = '  %s: %s' % (cls.sqlmeta.table, diff)
+                        all_diffs.append(diff)
+            if all_diffs:
+                print 'Database does not match schema:'
+                print '\n'.join(all_diffs)
+                for conn in conns:
+                    self.update_db(version, conn)
+        else:
+            all_diffs = []
         if self.options.open_editor:
             if not last_version_dir:
                 print ("Cannot edit upgrader because there is no "
@@ -952,7 +1066,7 @@ class CommandRecord(Command):
     def strip_comments(self, sql):
         lines = [l for l in sql.splitlines()
                  if not l.strip().startswith('--')]
-        return '\n'.join(lines)        
+        return '\n'.join(lines)
 
     def base_dir(self):
         base = self.options.output_dir
@@ -967,7 +1081,7 @@ class CommandRecord(Command):
     def find_output_dir(self):
         today = time.strftime('%Y-%m-%d', time.localtime())
         if self.options.version_name:
-            dir = os.path.join(self.base_dir(), today + '-' + 
+            dir = os.path.join(self.base_dir(), today + '-' +
                                self.options.version_name)
             if os.path.exists(dir):
                 print ("Error, directory already exists: %s"
@@ -983,7 +1097,7 @@ class CommandRecord(Command):
                 extra = 'a'
             else:
                 extra = chr(ord(extra)+1)
-    
+
     def find_last_version(self):
         names = []
         for fn in os.listdir(self.base_dir()):
@@ -1046,7 +1160,11 @@ class CommandUpgrade(CommandRecord):
         if self.options.upgrade_to:
             version_to = self.options.upgrade_to
         else:
-            version_to = os.path.basename(self.find_last_version())
+            fname = self.find_last_version()
+            if fname is None:
+                print "No version exists, use 'record' command to create one"
+                return
+            version_to = os.path.basename(fname)
         current = self.current_version()
         if v:
             print 'Current version: %s' % current
@@ -1076,7 +1194,7 @@ class CommandUpgrade(CommandRecord):
                     raise
             self.update_db(next_version, conn)
         print 'Done.'
-                
+
 
     def current_version(self):
         conn = self.connection()
@@ -1137,10 +1255,10 @@ class CommandUpgrade(CommandRecord):
                 print 'No upgraders found in %s' % current_dir
             return None, None
         upgraders.sort()
-        return upgraders[-1]        
-        
+        return upgraders[-1]
+
 def update_sys_path(paths, verbose):
-    if isinstance(paths, (str, unicode)):
+    if isinstance(paths, basestring):
         paths = [paths]
     for path in paths:
         path = os.path.abspath(path)
