@@ -3,32 +3,44 @@ import re
 from sqlobject import col
 from sqlobject import sqlbuilder
 from sqlobject.converters import registerConverter
-psycopg = None
-pgdb = None
 
 class PostgresConnection(DBAPI):
 
     supportTransactions = True
     dbName = 'postgres'
-    schemes = [dbName, 'postgresql', 'psycopg']
+    schemes = [dbName, 'postgresql']
 
     def __init__(self, dsn=None, host=None, port=None, db=None,
-                 user=None, password=None, usePygresql=False, unicodeCols=False,
+                 user=None, password=None, backend='psycopg', unicodeCols=False,
                  **kw):
-        global psycopg, pgdb
-        self.usePygresql = usePygresql
-        if usePygresql:
-            if pgdb is None:
-                import pgdb
-            self.module = pgdb
-        else:
-            if psycopg is None:
-                try:
+        backends = backend
+        for backend in backends.split(','):
+            backend = backend.strip()
+            if not backend:
+                continue
+            try:
+                if backend == 'psycopg2':
                     import psycopg2 as psycopg
-                except ImportError:
+                elif backend == 'psycopg1':
                     import psycopg
+                elif backend == 'psycopg':
+                    try:
+                        import psycopg2 as psycopg
+                    except ImportError:
+                        import psycopg
+                elif backend == 'pygresql':
+                    import pgdb
+                    self.module = pgdb
+                else:
+                    raise ValueError('Unknown PostgreSQL backend "%s", expected psycopg2, psycopg1 or pygresql' % backend)
+            except ImportError:
+                pass
+            else:
+                break
+        else:
+            raise ImportError('Cannot find a PostgreSQL backend, tried %s' % backends)
+        if backend.startswith('psycopg'):
             self.module = psycopg
-
             # Register a converter for psycopg Binary type.
             registerConverter(type(psycopg.Binary('')),
                               PsycoBinaryConverter)
@@ -42,7 +54,7 @@ class PostgresConnection(DBAPI):
         if host:
             dsn_dict["host"] = host
         if port:
-            if usePygresql:
+            if backend == 'pygresql':
                 dsn_dict["host"] = "%s:%d" % (host, port)
             else:
                 if psycopg.__version__.split('.')[0] == '1':
@@ -57,7 +69,7 @@ class PostgresConnection(DBAPI):
             dsn_dict["password"] = password
         self.use_dsn = dsn is not None
         if dsn is None:
-            if usePygresql:
+            if backend == 'pygresql':
                 dsn = ''
                 if host:
                     dsn += host
@@ -86,6 +98,10 @@ class PostgresConnection(DBAPI):
         self.dsn = dsn
         self.unicodeCols = unicodeCols
         self.schema = kw.pop('schema', None)
+        if "charset" in kw:
+            self.dbEncoding = kw.pop("charset")
+        else:
+            self.dbEncoding = None
         DBAPI.__init__(self, **kw)
 
     def connectionFromURI(cls, uri):
@@ -115,9 +131,12 @@ class PostgresConnection(DBAPI):
             # psycopg2 does not have an autocommit method.
             if hasattr(conn, 'autocommit'):
                 conn.autocommit(1)
+        c = conn.cursor()
         if self.schema:
-            c = conn.cursor()
             c.execute("SET search_path TO " + self.schema)
+        dbEncoding = self.dbEncoding
+        if dbEncoding:
+            c.execute("SET client_encoding TO %s" % dbEncoding)
         return conn
 
     def _queryInsertID(self, conn, soInstance, id, names, values):
@@ -306,7 +325,7 @@ class PostgresConnection(DBAPI):
         # We have to connect to *some* database, so we'll connect to
         # template1, which is a common open database.
         # @@: This doesn't use self.use_dsn or self.dsn_dict
-        if self.usePygresql:
+        if self.backend == 'pygresql':
             dsn = '%s:template1:%s:%s' % (
                 self.host or '', self.user or '', self.password or '')
         else:
